# encoding: utf-8

################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                            (C) 2020, Jeremy Brown                            #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from py_fast_trie.x_fast import TrieNode

def test_init():
	left = TrieNode("left", True)
	right = TrieNode("right", True)
	par = TrieNode("parent", False, left, right)

	assert left._value == "left"
	assert left._leaf == True
	assert left._left is None
	assert left._right is None
	assert left._parent is None

	assert right._value == "right"
	assert right._leaf == True
	assert right._left is None
	assert right._right is None
	assert right._parent is None

	assert par._value == "parent"
	assert par._leaf == False
	assert par._left is left
	assert par._right is right
	assert par._parent is None

def test_properties():
	left = TrieNode(0, True)
	right = TrieNode(1, True)
	middle = TrieNode(0, False, left, right)
	root = TrieNode(None, False, middle)
	
	left.parent = middle
	left.right = right

	right.parent = middle
	right.left = left

	middle.parent = root

	assert left.leaf == True
	assert left.left is None
	assert left.right is right
	assert left.parent is middle
	assert left.value == 0
	assert left.value_bits == "00"
	assert str(left) == "0"

	assert right.leaf == True
	assert right.left is left
	assert right.right is None
	assert right.parent is middle
	assert right.value == 1
	assert right.value_bits == "01"
	assert str(right) == "1"

	assert middle.leaf == False
	assert middle.left is left
	assert middle.right is right
	assert middle.parent is root
	assert middle.value is 0
	assert middle.value_bits == "0"
	assert str(middle) == "0"

	assert root.leaf == False
	assert root.left is middle
	assert root.right is None
	assert root.parent is None
	assert root.value is None
	assert root.value_bits == ""
	assert str(root) == "Root"



# encoding: utf-8

################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                            (C) 2020, Jeremy Brown                            #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from __future__ import division

from functools import partial
from sys import maxsize

from hypothesis import settings
from hypothesis.strategies import (binary,
								   integers,
								   lists,
								   none,
								   one_of,
								   )

from py_fast_trie import XFastTrie

max_trie_entry_size = maxsize.bit_length() + 1 if settings._current_profile == "ci" else 24
max_trie_value = (2 ** max_trie_entry_size) - 1

to_int = partial(XFastTrie._to_int, length=max_trie_entry_size)

invalid_binary_entry = binary(min_size=(max_trie_entry_size // 8 + 1))
invalid_int_entry = one_of(integers(max_value=-1),
						   integers(min_value=(max_trie_value + 2)))
invalid_trie_entry = one_of(invalid_binary_entry, invalid_int_entry, none())

valid_binary_entry = binary(min_size=1, max_size=(max_trie_entry_size // 8))
valid_int_entry = integers(min_value=0, max_value=max_trie_value)
valid_int_entries = lists(valid_int_entry, min_size=1, max_size=max_trie_value, unique=True)
valid_trie_entry = one_of(valid_binary_entry, valid_int_entry)
valid_trie_entries = lists(valid_trie_entry, min_size=1, max_size=max_trie_value, unique_by=to_int)



# encoding: utf-8

################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                            (C) 2020, Jeremy Brown                            #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from __future__ import division

from sys import maxsize

from sortedcontainers import SortedList

from py_fast_trie import XFastTrie
from py_hopscotch_dict import HopscotchDict

class YFastTrie(object):

	@staticmethod
	def _calculate_representative(value, max_length):
		"""
		Calculate the smallest value that would exist in _partitions
		the given value could belong to

		:param val: The value to calculate the representative for
		:param max_length: The bit length of the largest possible representative

		:return: (int) The closest possible representative to the given number
		"""
		return min(max_length * (value // max_length) + (-1 % max_length), 2 ** max_length - 1)

	@staticmethod
	def _merge_subtrees(left_tree, right_tree, max_size):
		"""
		Combine the elements of two trees into one larger tree,
		splitting them again if the larger tree exceeds a given size

		:param left_tree: Tree containing smaller elements
		:param right_tree: Tree containing larger elements
		:param max_size: Maximum size the combined tree can be before splitting
		:return: (tuple) The combined tree and None if both trees' elements are
						 less than max_size, the tree with the smaller elements
						 and the tree with the larger elements otherwise
		"""
		if len(left_tree) + len(right_tree) <= max_size:
			left_tree.update(right_tree)
			result = (left_tree, None)
		else:
			total_median = (len(left_tree) + len(right_tree)) // 2 - 1
			if total_median < len(left_tree):
				total_median = left_tree[total_median]
			else:
				total_median = right_tree[total_median - len(left_tree)]

			# Taking advantage of the fact the number of elements in a subtree can only
			# be twice the bit length of the maximum element before splitting
			median_rep = YFastTrie._calculate_representative(total_median, max_size // 2)

			if median_rep <= max(left_tree):
				from_tree = left_tree
				to_tree = right_tree
				side = -1
			else:
				from_tree = right_tree
				to_tree = left_tree
				side = 0

			while max(left_tree) > median_rep or min(right_tree) <= median_rep:
				to_tree.add(from_tree.pop(side))

			result = (left_tree, right_tree)

		return result

	@staticmethod
	def _split_subtree(tree, max_length):
		"""
		Split a tree by its median element into two smaller trees

		:param tree: The tree to split
		:param max_length: The size of the largest possible element in the trie in bits
		:return: (tuple) The tree with the smaller elements,
						 and the tree with the larger elements
		"""
		median = tree.bisect_right(YFastTrie._calculate_representative(tree[len(tree) // 2], max_length))
		return SortedList(tree.islice(stop=median)), SortedList(tree.islice(start=median))

	def _get_value_subtree(self, value, create_subtree=False):
		"""
		Find the subtree that would hold the given value

		:param value: The value to find
		:param create_subtree: If there is no subtree that would hold the given value,
							   create one
		:return: (tuple) The subtree that potentially holds the given value,
						 and its corresponding representative
		"""
		result = None

		if self._count == 0:
			rep_node = None
		elif value <= self._min or self._min is None:
			rep_node = self._partitions.min_node
		else:
			# As the X-fast trie looks for strict successors,
			# if the value being searched for is a representative,
			# the wrong representative will be returned if the one being searched for
			# is not the largest, and no representative will be returned at all if it is;
			# so subtract one before searching for the successor
			rep_node = self._partitions.successor(value - 1)

		if rep_node is None:
			if create_subtree:
				rep = self._calculate_representative(value, self._maxlen)
				self._partitions += rep
				rep_node = self._partitions.successor(rep - 1)
				self._subtrees[rep] = result = SortedList()
		else:
			# Every representative in the X-fast trie should have a corresponding SortedList;
			# the code should blow up if it doesn't
			result = self._subtrees[rep_node.value]

		return (result, rep_node)

	def clear(self):
		"""
		Remove all values from the trie and return it to its starting state
		"""
		self._count = 0
		self._max = None
		self._min = None
		self._partitions = XFastTrie(self._maxlen)
		self._subtrees = HopscotchDict()

	def insert(self, value):
		"""
		Insert a value into the trie

		:param value: The value to insert into the trie
		"""
		value = XFastTrie._to_int(value, self._maxlen)
		subtree, rep_node = self._get_value_subtree(value, True)
		# Do nothing if the value is already in the trie
		if value in subtree:
			return

		if self._max is None or value > self._max:
			self._max = value

		if self._min is None or value < self._min:
			self._min = value

		subtree.add(value)

		if len(subtree) > self._max_subtree_size:
			# Out with the old
			del self._subtrees[rep_node.value]
			self._partitions -= rep_node.value

			# In with the new
			for tree in self._split_subtree(subtree, self._maxlen):
				rep = self._calculate_representative(max(tree), self._maxlen)
				self._partitions += rep
				self._subtrees[rep] = tree

		self._count += 1

	def predecessor(self, value):
		"""
		Find the largest value in the trie strictly less than the given value,
		if it exists

		:param value: The value to find the predecessor of
		:return: (int) The predecessor of the given value, or None if it doesn't exist
		"""
		value = XFastTrie._to_int(value, self._maxlen)
		subtree, rep_node = self._get_value_subtree(value)

		# subtree should be None only if the trie is empty
		if subtree is None and self._count == 0:
			raise ValueError(u"No values exist in trie")
		elif value <= self._min or self._min is None:
			return None
		elif value > self._max:
			return self._max

		if min(subtree) >= value:
			subtree = self._subtrees[rep_node.pred.value]

		return subtree[subtree.bisect_left(value) - 1]

	def remove(self, value):
		"""
		Remove the given value from the trie

		:param value: The value to remove from the trie
		"""
		value = XFastTrie._to_int(value, self._maxlen)
		subtree, rep_node = self._get_value_subtree(value)

		# There should be no subtree only if the given value is not in the trie
		if subtree is None or value not in subtree:
			raise ValueError(u"Value does not exist in trie")

		if self._min == value:
			if len(subtree) > 1:
				min_succ = subtree[1]
			else:
				min_succ = self.successor(value)
		else:
			min_succ = -1

		if self._max == value:
			if len(subtree) > 1:
				max_pred = subtree[-2]
			else:
				max_pred = self.predecessor(value)
		else:
			max_pred = -1

		if min_succ != -1:
			self._min = min_succ

		if max_pred != -1:
			self._max = max_pred

		subtree.remove(value)

		if len(subtree) == 0:
			del self._subtrees[rep_node.value]
			self._partitions -= rep_node.value

		elif len(subtree) < self._min_subtree_size and len(self._partitions) > 1:
			if rep_node.pred is not None:
				left_rep = rep_node.pred
				right_rep = rep_node
			else:
				left_rep = rep_node
				right_rep = rep_node.succ

			left_tree = self._subtrees[left_rep.value]
			right_tree = self._subtrees[right_rep.value]

			# Out with the old
			del self._subtrees[left_rep.value]
			del self._subtrees[right_rep.value]
			self._partitions -= left_rep.value
			self._partitions -= right_rep.value

			# In with the new
			for tree in filter(None, self._merge_subtrees(left_tree, right_tree, 2 * self._maxlen)):
				rep = self._calculate_representative(max(tree), self._maxlen)
				self._partitions += rep
				self._subtrees[rep] = tree

		self._count -= 1

	def successor(self, value):
		"""
		Find the smallest value in the trie strictly greater than the given value,
		if it exists

		:param value: The value to find the successor of
		:return: (int) The successor of the given value, or None if it doesn't exist
		"""
		value = XFastTrie._to_int(value, self._maxlen)
		subtree, rep_node = self._get_value_subtree(value)

		# subtree should be None only if the trie is empty
		if subtree is None and self._count == 0:
			raise ValueError(u"No values exist in trie")
		elif value >= self._max or self._max is None:
			return None
		elif value < self._min:
			return self._min

		if max(subtree) <= value:
			subtree = self._subtrees[rep_node.succ.value]

		return subtree[subtree.bisect_right(value)]

	@property
	def max(self):
		"""
		The maximum value in the trie

		:return: (int) The maximum value in the trie,
					   or None if the trie is empty
		"""
		return self._max

	@property
	def min(self):
		"""
		The minimum value in the trie

		:return: (int) The minimum value in the trie,
					   or None if the trie is empty
		"""
		return self._min

	def __init__(self, max_length=(maxsize.bit_length() + 1)):
		self._maxlen = max_length
		self._min_subtree_size = max_length // 2
		self._max_subtree_size = max_length * 2
		self.clear()

	def __contains__(self, value):
		value = XFastTrie._to_int(value, self._maxlen)
		subtree, _ = self._get_value_subtree(value)
		return subtree is not None and value in subtree

	def __gt__(self, value):
		value = XFastTrie._to_int(value, self._maxlen)
		return self.successor(value)

	def __iadd__(self, value):
		value = XFastTrie._to_int(value, self._maxlen)
		self.insert(value)
		return self

	def __isub__(self, value):
		value = XFastTrie._to_int(value, self._maxlen)
		self.remove(value)
		return self

	def __iter__(self):
		for rep in sorted(self._subtrees):
			for value in self._subtrees[rep]:
				yield value

	def __len__(self):
		return self._count

	def __lt__(self, value):
		value = XFastTrie._to_int(value, self._maxlen)
		return self.predecessor(value)



# encoding: utf-8

################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                            (C) 2020, Jeremy Brown                            #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from hypothesis import HealthCheck, settings
from hypothesis.database import ExampleDatabase

settings.register_profile("ci",
						  database=ExampleDatabase(":memory:"),
						  max_examples=500,
						  stateful_step_count=200,
						  suppress_health_check=[HealthCheck.too_slow])



# encoding: utf-8

################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                            (C) 2020, Jeremy Brown                            #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from io import open
from os.path import abspath, dirname, join
from setuptools import find_packages, setup


package_root = abspath(dirname(__file__))

# Get the long description from the README file
with open(join(package_root, "README.md"), encoding="utf-8") as desc:
	long_description = desc.read()


setup(
	name="py-fast-trie",

	packages=find_packages(where="src"),

	package_dir={"": "src"},

	license="NPOSL-3.0",

	url="https://github.com/mischif/py-fast-trie",

	description="Python library for tries with different grades of fastness",

	long_description=long_description,
	long_description_content_type="text/markdown",

	author="Jeremy Brown",
	author_email="mischif@users.noreply.github.com",

	python_requires="~=3.6",

	package_data={"py_fast_trie": ["VERSION"]},

	install_requires=["py-hopscotch-dict", "sortedcontainers"],
	
	setup_requires=["pytest-runner", "setuptools_scm"],

	tests_require=["hypothesis", "hypothesis-pytest", "pytest", "pytest-cov"],

	zip_safe=False,

	keywords=["x-fast", "y-fast", "trie", "data structures"],

	extras_require={
		"test": ["codecov"],
		},

	options={
		"aliases": {
			"test": "pytest",
			},

		"metadata": {
			"license_files": "LICENSE",
			},
		},

	classifiers=[
		"Development Status :: 5 - Production/Stable",

		"Operating System :: OS Independent",

		"License :: OSI Approved :: Open Software License 3.0 (OSL-3.0)",

		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",

		"Topic :: Software Development :: Libraries",
		"Topic :: Software Development :: Libraries :: Python Modules",
		],
	)



# encoding: utf-8

################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                            (C) 2020, Jeremy Brown                            #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from __future__ import division

from numbers import Integral
from sys import maxsize

from py_hopscotch_dict import HopscotchDict

try:
	from future_builtins import map
except ImportError:
	pass


class TrieNode(object):
	def _get_leaf(self):
		"""
		Indicated whether or not the node is a leaf

		:return: (bool) Whether or not the node is a leaf
		"""
		return self._leaf

	def _get_left(self):
		"""
		The left child of the current node -
		its predecessor if the current node is a leaf,
		its left child if the current node is an internal node with such,
		a descendant pointer if the current node is an internal node with no child

		:return: (TrieNode) The current node's left child,
							None if the current node is the smallest leaf
		"""
		return self._left

	def _get_parent(self):
		"""
		The parent of the current node

		:return: (TrieNode) The current node's parent,
							None if the current node is the root
		"""
		return self._parent

	def _get_right(self):
		"""
		The right child of the current node -
		its successor if the current node is a leaf,
		its right child if the current node is an internal node with such,
		a descendant pointer if the current node is an internal node with no child

		:return: (TrieNode) The current node's right child,
							None if the current node is the largest leaf
		"""
		return self._right

	def _get_value(self):
		"""
		The value of the current node, expressed as an integer

		:return: (int) The current node's value
		"""
		return self._value

	def _get_value_bitstring(self):
		"""
		The value of the current node, expressed as a string of 1s and 0s

		:return: (str) The current node's value, as bits
		"""
		node = self
		result = []

		while node.value is not None:
			result.append(str(node.value & 1))
			node = node.parent

		return "".join(reversed(result))

	def _set_left(self, new_left):
		"""
		Sets the left child of the current node

		:param new_left: (TrieNode) The current node's left child, or None
		"""
		self._left = new_left

	def _set_parent(self, new_parent):
		"""
		Sets the parent of the current node

		:param new_parent: (TrieNode) The current node's parent, or None
		"""
		self._parent = new_parent

	def _set_right(self, new_right):
		"""
		Sets the left child of the current node

		:param new_right: (TrieNode) The current node's right child, or none
		"""
		self._right = new_right

	leaf = property(_get_leaf)
	value = property(_get_value)
	value_bits = property(_get_value_bitstring)
	parent = property(_get_parent, _set_parent)
	left = property(_get_left, _set_left)
	right = property(_get_right, _set_right)
	pred = property(_get_left, _set_left)
	succ = property(_get_right, _set_right)

	def __init__(self, value, leaf, left=None, right=None):
		self._leaf = leaf
		self._value = value
		self._left = left
		self._right = right
		self._parent = None

	def __str__(self):
		return "Root" if self._value is None else str(self._value)


class XFastTrie(object):
	@staticmethod
	def _make_level_tables(levels):
		"""
		Creates the dicts used when searching for a value in the trie

		:param levels: (int) The number of levels in the trie
		:return: (list) search structures for each level of the trie
		"""
		return [HopscotchDict() for _ in range(levels)]

	@staticmethod
	def _to_int(value, length):
		"""
		Confirm the desired value could be contained in the table,
		then perform any necessary conversions to the canonical value format

		:param value: (int/bytes) The value to be converted
		:param length: The maximum bit length of a value in the trie
		:return: (int) The value converted to an int
		"""
		if isinstance(value, Integral):
			if value.bit_length() > length:
				raise ValueError(u"Value is too big to be stored in trie")
			elif value < 0:
				raise ValueError(u"Negative values cannot be stored in trie")
			else:
				return value

		elif isinstance(value, bytes):
			if len(value) * 8 > length:
				raise ValueError(u"Value is too big to be stored in trie")

			else:
				return sum(map(lambda t: t[1] << 8 * t[0], enumerate(reversed(value))))

		else:
			raise TypeError(u"Only integers and byte sequences can be stored in trie")

	def _get_closest_ancestor(self, value):
		"""
		Find the node in the trie with the longest prefix that matches the given value

		:param value: (int) The value to search for
		:return: (TrieNode, int) The node with the longest prefix matching the given value,
								 and its depth in the trie
		"""
		result = self._root
		result_level = -1

		low_side = 0
		high_side = self._maxlen - 1

		while low_side <= high_side:
			level = (low_side + high_side) // 2
			prefix = value >> (self._maxlen - level - 1)

			if prefix not in self._level_tables[level]:
				high_side = level - 1
			else:
				result = self._level_tables[level][prefix]
				result_level = level
				low_side = level + 1

		return (result, result_level)

	def _get_closest_leaf(self, value):
		"""
		Find the leaf in the trie with the value closest to the given value

		:param value: (int) The value to search for
		:return: (TrieNode) The leaf with the closest value to the given value
		"""
		result = None
		ancestor, level = self._get_closest_ancestor(value)

		# The value is stored in the trie and therefore is the closest leaf to itself
		if ancestor.leaf:
			return ancestor

		# The pointer to the next level down is a descendant pointer
		# and may be larger or smaller than the given value, depending on the leg
		else:
			direction = value >> (self._maxlen - level - 2) & 1
			descendant = ancestor.left if direction == 0 else ancestor.right

			if descendant is not None:
				# The descendant pointer only points to values on the other leg of the ancestor;
				# make sure there is no leaf not a child of the ancestor which is closer
				candidate = descendant.pred if direction == 0 else descendant.succ

				if candidate is None or abs(descendant.value - value) < abs(candidate.value - value):
					result =  descendant
				else:
					result = candidate
		return result

	def clear(self):
		"""
		Empty the trie of all values
		"""
		self._level_tables = self._make_level_tables(self._maxlen)
		self._root = TrieNode(None, False)
		self._count = 0
		self._min = None
		self._max = None

	def insert(self, value):
		"""
		Add the given value to the trie

		:param value: (int/bytes) The value to add to the trie
		"""
		value = self._to_int(value, self._maxlen)

		# Do nothing if the value is already in the trie
		if value in self._level_tables[-1]:
			return

		leaf_pred = self.predecessor(value) if self._count > 0 else None
		leaf_succ = self.successor(value) if self._count > 0 else None
		leaf_node = TrieNode(value, True, leaf_pred, leaf_succ)

		# Wire the new leaf into the linked list and add to the leaf dict
		self._level_tables[-1][value] = leaf_node

		if leaf_pred is not None:
			leaf_pred.succ = leaf_node

		if leaf_succ is not None:
			leaf_succ.pred = leaf_node

		# Update global min/max pointers as necessary
		if self._min is None or value < self._min.value:
			self._min = leaf_node

		if self._max is None or value > self._max.value:
			self._max = leaf_node

		# Walk up the trie from the leaf node, creating internal nodes as necessary
		last_inserted = leaf_node
		for level in reversed(range(self._maxlen - 1)):
			node_value = int(value >> (self._maxlen - level - 1))
			node = self._level_tables[level].get(node_value)

			if node is None:
				# Determine which leg the last node inserted into the trie was on relative the one to be created,
				# and find the corresponding leaf to use for the descendant pointer
				last_inserted_leg = last_inserted.value & 1
				descendant_direction = "right" if last_inserted_leg == 0 else "left"
				descendant = last_inserted
				while not descendant.leaf:
					# If this loop ends up following a descendant pointer,
					# it means there was no intermediate node to follow instead;
					# a pointer on the left leg would lead to the smallest leaf of the node's right subtree,
					# which would also be the smallest leaf of the original node and the desired node,
					# and likewise for a descendant pointer on the right leg.
					descendant = getattr(descendant, descendant_direction)

				if last_inserted_leg == 0:
					node_left = last_inserted
					node_right = descendant
				else:
					node_right = last_inserted
					node_left = descendant

				# Create the new node, insert it into its respective dict and update pointers
				node = TrieNode(node_value, False, node_left, node_right)
				self._level_tables[level][node_value] = node
				last_inserted.parent = node
				last_inserted = node
			else:
				# Update the last node that was inserted to point to its newly-found parent
				if last_inserted.parent is None:
					last_inserted.parent = node

				# Update the current node to know about its potential children
				left_child_value = node.value << 1 & -2
				right_child_value = node.value << 1 | 1
				left_child = self._level_tables[level + 1].get(left_child_value)
				right_child = self._level_tables[level + 1].get(right_child_value)

				if node.left is left_child:
					pass
				elif left_child is not None:
					node.left = left_child
				else:
					if node.left.value > value:
						node.left = leaf_node

				if node.right is right_child:
					pass
				elif right_child is not None:
					node.right = right_child
				else:
					if node.right.value < value:
						node.right = leaf_node

		if self._root.left is None or self._root.left.leaf:
			root_left = self._level_tables[0].get(0)
			self._root.left = root_left or self._min
			if root_left is not None:
				root_left.parent = self._root

		if self._root.right is None or self._root.right.leaf:
			root_right = self._level_tables[0].get(1)
			self._root.right = root_right or self._max
			if root_right is not None:
				root_right.parent = self._root

		self._count += 1

	def predecessor(self, value):
		"""
		Find the largest value in the trie strictly less than the given value

		:param value: (int) The value to find the predecessor for
		:return: (TrieNode) The leaf with the largest value strictly less than the given value,
							or None if the value is at most the value of the smallest leaf
		"""
		value = self._to_int(value, self._maxlen)
		node = self._get_closest_leaf(value)

		# This should only happen if there are no values in the trie,
		# But if it could also happen because of some unconsidered edge case,
		# make some noise so the edge case can be fixed
		if node is None:
			raise ValueError(u"No values exist in trie")
		else:
			return node.pred if node.value >= value else node

	def remove(self, value):
		"""
		Remove the given value from the trie

		:param value: (int/bytes) The value to remove from the trie
		"""
		value = self._to_int(value, self._maxlen)

		# Error when trying to remove a value that hasn't been added
		if value not in self._level_tables[-1]:
			raise ValueError(u"Value does not exist in trie")
		else:
			node = self._level_tables[-1][value]
			leaf_pred = node.pred
			leaf_succ = node.succ

			# Take the value out of the leaf dict and linked list
			del self._level_tables[-1][value]

			if leaf_pred is not None:
				leaf_pred.succ = leaf_succ

			if leaf_succ is not None:
				leaf_succ.pred = leaf_pred

			# Update global min/max pointers as necessary
			if self._min is node:
				self._min = leaf_succ

			if self._max is node:
				self._max = leaf_pred

			# Walk up the trie from the leaf node, modifying/removing internal nodes as necessary
			node = node.parent
			level = self._maxlen - 2
			while node.value is not None:
				left_child_value = node.value << 1 & -2
				right_child_value = node.value << 1 | 1
				left_child = self._level_tables[level + 1].get(left_child_value)
				right_child = self._level_tables[level + 1].get(right_child_value)

				# Node has no children, delete it
				if left_child is None and right_child is None:
					del self._level_tables[level][node.value]

				else:
					descendant = left_child if right_child is None else right_child
					leg = "right" if right_child is None else "left"
					while not descendant.leaf:
						# If this loop ends up following a descendant pointer,
						# it means there was no intermediate node to follow instead;
						# a pointer on the left leg would lead to the smallest leaf of the node's right subtree,
						# which would also be the smallest leaf of the original node and the desired node,
						# and likewise for a descendant pointer on the right leg.
						descendant = getattr(descendant, leg)

					if left_child is None:
						node.left = descendant

					elif right_child is None:
						node.right = descendant

				node = node.parent
				level -= 1

		root_left = self._level_tables[0].get(0)
		root_right = self._level_tables[0].get(1)

		# Update root
		if root_left is None and root_right is None:
			self._root.left = self._root.right = None

		elif root_left is None:
			self._root.left = self._min

		elif root_right is None:
			self._root.right = self._max

		self._count -= 1

	def successor(self, value):
		"""
		Find the smallest value in the trie strictly greater than the given value

		:param value: (int) The value to find the successor for
		:return: (TrieNode) The leaf with the smallest value strictly greater than the given value,
							or None if the value is at least the value of the largest leaf
		"""
		value = self._to_int(value, self._maxlen)
		node = self._get_closest_leaf(value)

		# This should only happen if there are no values in the trie,
		# But if it could also happen because of some unconsidered edge case,
		# make some noise so the edge case can be fixed
		if node is None:
			raise ValueError(u"No values exist in trie")
		else:
			return node.succ if node.value <= value else node

	@property
	def max(self):
		"""
		The maximum value in the trie

		:return: (int) The maximum value in the trie,
					   or None if the trie is empty
		"""
		return self._max.value

	@property
	def max_node(self):
		"""
		The node related to the maximum value in the trie

		:return: (TrieNode) The maximum value in the trie,
							or None if the trie is empty
		"""
		return self._max

	@property
	def min(self):
		"""
		The minimum value in the trie

		:return: (int) The minimum value in the trie,
					   or None if the trie is empty
		"""
		return self._min.value

	@property
	def min_node(self):
		"""
		The node related to the minimum value in the trie

		:return: (TrieNode) The minimum value in the trie,
							or None if the trie is empty
		"""
		return self._min

	def __init__(self, max_length=(maxsize.bit_length() + 1)):
		self._maxlen = max_length
		self.clear()

	def __contains__(self, value):
		value = self._to_int(value, self._maxlen)
		return value in self._level_tables[-1]

	def __gt__(self, value):
		value = self._to_int(value, self._maxlen)
		result = self.successor(value)
		return result.value if result is not None else result

	def __iadd__(self, value):
		value = self._to_int(value, self._maxlen)
		self.insert(value)
		return self

	def __isub__(self, value):
		value = self._to_int(value, self._maxlen)
		self.remove(value)
		return self

	def __iter__(self):
		node = self._min
		while node is not None:
			yield node.value
			node = node.succ

	def __len__(self):
		return self._count

	def __lt__(self, value):
		value = self._to_int(value, self._maxlen)
		result = self.predecessor(value)
		return result.value if result is not None else result



# encoding: utf-8

################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                            (C) 2020, Jeremy Brown                            #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from __future__ import division

from numbers import Integral
from struct import pack, unpack
from sys import maxsize

import pytest

from hypothesis import given, note, seed, settings
from hypothesis.strategies import integers, lists
from hypothesis.stateful import RuleBasedStateMachine, invariant, rule

from py_fast_trie import XFastTrie
from test import (invalid_trie_entry,
				  max_trie_entry_size,
				  max_trie_value,
				  valid_int_entries,
				  valid_int_entry,
				  valid_trie_entries,
				  valid_trie_entry)


def to_bytes(val):
	if val.bit_length() < 9:
		fmt = "B"
	elif val.bit_length() < 17:
		fmt = ">H"
	elif val.bit_length() < 33:
		fmt = ">L"
	else:
		fmt = ">Q"

	return pack(fmt, val)


@given(integers(min_value=0, max_value=max_trie_entry_size))
def test_make_level_tables(depth):
	assert len(XFastTrie._make_level_tables(depth)) == depth


@given(valid_trie_entry)
def test_to_int(value):
	if isinstance(value, int):
		assert XFastTrie._to_int(value, max_trie_entry_size) == value

	elif isinstance(value, bytes):
		value_int = unpack(">Q", value.rjust((maxsize.bit_length() + 1) // 8, b'\x00'))[0]
		assert XFastTrie._to_int(value, max_trie_entry_size) == value_int


@given(invalid_trie_entry)
def test_to_int_exceptions(value):
	if isinstance(value, Integral):
		with pytest.raises(ValueError):
			XFastTrie._to_int(value, max_trie_entry_size)

	elif isinstance(value, bytes):
		with pytest.raises(ValueError):
			XFastTrie._to_int(value, max_trie_entry_size)

	else:
		with pytest.raises(TypeError):
			XFastTrie._to_int(value, max_trie_entry_size)


@given(valid_trie_entries, valid_int_entries)
def test_get_closest_ancestor(entries, test_values):
	t = XFastTrie(max_trie_entry_size)

	for entry in entries:
		t += entry

	entries = [t._to_int(e, t._maxlen) for e in entries]

	for val in test_values:
		ancestor, level = t._get_closest_ancestor(val)

		if val in entries:
			assert ancestor.leaf
			assert ancestor.value == val

		else:
			test_bits = format(val, 'b').zfill(t._maxlen)[:level + 2]
			assert not ancestor.leaf
			assert not ancestor.left.value_bits.startswith(test_bits)
			assert not ancestor.right.value_bits.startswith(test_bits)


@given(valid_trie_entries, valid_int_entries)
def test_get_closest_leaf(entries, test_values):
	t = XFastTrie(max_trie_entry_size)

	for entry in entries:
		t += entry

	entries = [t._to_int(e, t._maxlen) for e in entries]

	for val in test_values:
		neighbor = t._get_closest_leaf(val)
		assert neighbor.leaf

		if val in entries:
			assert neighbor.value == val

		else:
			if neighbor.pred is not None:
				assert abs(neighbor.value - val) <= abs(neighbor.pred.value - val)

			if neighbor.succ is not None:
				assert abs(neighbor.value - val) <= abs(neighbor.succ.value - val)


@given(valid_trie_entries, valid_int_entries)
def test_predecessor(entries, test_values):
	t = XFastTrie(max_trie_entry_size)

	for entry in entries:
		t += entry

	for val in test_values:
		pred = t < val

		if pred is not None:
			assert pred < val
			pred = t.predecessor(val)

			if pred.succ is not None:
				assert pred.succ.value >= val


@given(valid_trie_entries, valid_int_entries)
def test_successor(entries, test_values):
	t = XFastTrie(max_trie_entry_size)

	for entry in entries:
		t += entry

	for val in test_values:
		succ = t > val

		if succ is not None:
			assert succ > val
			succ = t.successor(val)

			if succ.pred is not None:
				assert succ.pred.value <= val


def test_successor_predecessor_empty_trie():
	t = XFastTrie(max_trie_entry_size)

	with pytest.raises(ValueError):
		t.successor(0)

	with pytest.raises(ValueError):
		t.predecessor(0)


@given(valid_trie_entries)
def test_clear(entries):
	t = XFastTrie(max_trie_entry_size)

	for entry in entries:
		t += entry

	assert len(t) > 0
	assert t.min_node is not None
	assert t.max_node is not None

	t.clear()

	for d in t._level_tables:
		assert len(d) == 0

	assert len(t) == 0
	assert t.min_node is None
	assert t.max_node is None


@given(lists(valid_int_entry, min_size=0, max_size=max_trie_value, unique=True))
def test_iter(entries):
	t = XFastTrie(max_trie_entry_size)

	for entry in entries:
		t += entry

	entries = sorted(entries)

	for entry in t:
		assert entry == entries.pop(0)

	assert len(entries) == 0


class XFastStateMachine(RuleBasedStateMachine):
	def __init__(self):
		super(XFastStateMachine, self).__init__()
		self.t = XFastTrie(max_trie_entry_size)

	def teardown(self):
		values = list(self.t._level_tables[-1])

		for val in values:
			self.t -= val

	@invariant()
	def valid_count(self):
		assert len(self.t) == len(self.t._level_tables[-1])

	@invariant()
	def valid_min(self):
		if len(self.t) > 0:
			assert self.t.min_node.pred is None

			for leaf in self.t._level_tables[-1].values():
				assert leaf is self.t.min_node or self.t.min < leaf.value

	@invariant()
	def valid_max(self):
		if len(self.t) > 0:
			assert self.t.max_node.succ is None

			for leaf in self.t._level_tables[-1].values():
				assert leaf is self.t.max_node or self.t.max > leaf.value

	@invariant()
	def valid_pointers(self):
		for (level, table) in enumerate(self.t._level_tables):
			for node in table.values():
				if not node.leaf:
					left_child_value = node.value << 1 & -2
					right_child_value = node.value << 1 | 1
					left_child = self.t._level_tables[level + 1].get(left_child_value)
					right_child = self.t._level_tables[level + 1].get(right_child_value)

					# While we're in here, make sure the node should be in the trie
					assert left_child is not None or right_child is not None

					if left_child is not None:
						assert node.left is left_child
						assert left_child.parent is node
					else:
						desc = node.right

						while not desc.leaf:
							desc = desc.left

						assert node.left is desc

					if right_child is not None:
						assert node.right is right_child
						assert right_child.parent is node
					else:
						desc = node.left

						while not desc.leaf:
							desc = desc.right

						assert node.right is desc
				else:
					if node.pred is not None:
						assert node.pred.value < node.value

					if node.succ is not None:
						assert node.succ.value > node.value

	@rule(val=valid_trie_entry)
	def insert_value(self, val):
		self.t += val

	@rule(val=valid_trie_entry)
	def remove_value(self, val):
		if val not in self.t:
			try:
				self.t -= val
			except ValueError:
				pass
		else:
			self.t -= val

XFastStateMachine.TestCase.settings = settings(max_examples=50, deadline=None)
test_x_fast_trie = XFastStateMachine.TestCase



# encoding: utf-8

################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                            (C) 2020, Jeremy Brown                            #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from __future__ import division

from itertools import chain
from random import randint

import pytest

from hypothesis import given, settings
from hypothesis.strategies import integers, lists
from hypothesis.stateful import RuleBasedStateMachine, invariant, rule
from sortedcontainers import SortedList

from py_fast_trie import YFastTrie
from test import (invalid_trie_entry,
				  max_trie_entry_size,
				  max_trie_value,
				  valid_int_entries,
				  valid_int_entry,
				  valid_trie_entries,
				  valid_trie_entry,
				  XFastTrie
				  )


@given(integers(min_value=0, max_value=max_trie_value))
def test_calculate_representative(val):
	rep = YFastTrie._calculate_representative(val, max_trie_entry_size)
	assert rep >= val
	assert (rep + 1) % max_trie_entry_size == 0


@given(lists(valid_int_entry, min_size=2, max_size=((4 * max_trie_entry_size) - 1), unique=True))
def test_merge_subtrees(values):
	values = SortedList(values)
	split = randint(1, len(values) - 1)
	left_tree = SortedList(values.islice(stop=split))
	right_tree = SortedList(values.islice(start=split))
	new_left, new_right = YFastTrie._merge_subtrees(left_tree, right_tree, 2 * max_trie_entry_size)

	if len(values) <= 2 * max_trie_entry_size:
		assert new_right is None
		assert isinstance(new_left, SortedList)
		assert len(new_left) == len(values)
	else:
		assert isinstance(new_left, SortedList)
		assert isinstance(new_right, SortedList)
		assert len(new_left) + len(new_right) == len(values)
		assert YFastTrie._calculate_representative(max(new_left), max_trie_entry_size) < min(new_right)


@given(lists(valid_int_entry, min_size=((2 * max_trie_entry_size) + 1), max_size=((4 * max_trie_entry_size) - 1), unique=True))
def test_merge_large_subtrees(values):
	values = SortedList(values)
	split = len(values) // 2
	left_tree = SortedList(values.islice(stop=split))
	right_tree = SortedList(values.islice(start=split))
	new_left, new_right = YFastTrie._merge_subtrees(left_tree, right_tree, 2 * max_trie_entry_size)
	assert isinstance(new_left, SortedList)
	assert isinstance(new_right, SortedList)
	assert len(new_left) + len(new_right) == len(values)
	assert YFastTrie._calculate_representative(max(new_left), max_trie_entry_size) < min(new_right)

	split += 1
	left_tree = SortedList(values.islice(stop=split))
	right_tree = SortedList(values.islice(start=split))
	new_left, new_right = YFastTrie._merge_subtrees(left_tree, right_tree, 2 * max_trie_entry_size)
	assert isinstance(new_left, SortedList)
	assert isinstance(new_right, SortedList)
	assert len(new_left) + len(new_right) == len(values)
	assert YFastTrie._calculate_representative(max(new_left), max_trie_entry_size) < min(new_right)


@given(lists(valid_int_entry, min_size=((2 * max_trie_entry_size) + 1), max_size=((4 * max_trie_entry_size) - 1), unique=True))
def test_split_subtree(values):
	left_tree, right_tree = YFastTrie._split_subtree(SortedList(values), max_trie_entry_size)

	assert isinstance(left_tree, SortedList)
	assert isinstance(right_tree, SortedList)
	assert len(left_tree) + len(right_tree) == len(values)
	assert max(left_tree) < min(right_tree)
	assert YFastTrie._calculate_representative(max(left_tree), max_trie_entry_size) < min(right_tree)


@given(valid_trie_entries, valid_int_entries)
def test_get_value_subtree(entries, test_values):
	t = YFastTrie(max_trie_entry_size)

	assert t._get_value_subtree(test_values[0]) == (None, None)

	for entry in entries:
		t += entry

	entries = [XFastTrie._to_int(e, t._maxlen) for e in entries]

	for val in test_values:
		tree, rep_node = t._get_value_subtree(val)

		if val > t._partitions.max:
			assert tree is None
			assert rep_node is None

		else:
			assert val <= rep_node.value

			if val in entries:
				assert val in tree
			else:
				assert val not in tree


@given(valid_trie_entries, valid_int_entries)
def test_predecessor(entries, test_values):
	t = YFastTrie(max_trie_entry_size)

	for entry in entries:
		t += entry

	for val in test_values:
		pred = t < val

		if pred is not None:
			assert pred < val

			tree, _ = t._get_value_subtree(val)
			if tree is None or min(tree) >= val:
				tree, _ = t._get_value_subtree(pred)

			assert max([v for v in tree if v < val]) == pred

		else:
			assert val <= t.min


@given(valid_trie_entries, valid_int_entries)
def test_successor(entries, test_values):
	t = YFastTrie(max_trie_entry_size)

	for entry in entries:
		t += entry

	for val in test_values:
		succ = t > val

		if succ is not None:
			assert succ > val

			tree, _ = t._get_value_subtree(val)
			if tree is None or max(tree) <= val:
				tree, _ = t._get_value_subtree(succ)

			assert min([v for v in tree if v > val]) == succ

		else:
			assert val >= t.max


def test_successor_predecessor_empty_trie():
	t = YFastTrie(max_trie_entry_size)

	with pytest.raises(ValueError):
		t.successor(0)

	with pytest.raises(ValueError):
		t.predecessor(0)


@given(valid_trie_entries)
def test_clear(entries):
	t = YFastTrie(max_trie_entry_size)

	for entry in entries:
		t += entry

	assert len(t) > 0
	assert t.min is not None
	assert t.max is not None

	t.clear()

	assert len(t) == 0
	assert len(t._partitions) == 0
	assert len(t._subtrees) == 0
	assert t.min is None
	assert t.max is None


def test_insert_with_split():
	t = YFastTrie(max_trie_entry_size)

	big_rep = 3 * max_trie_entry_size - 1
	small_rep = 2 * max_trie_entry_size - 1
	for i in range(3 * max_trie_entry_size - 1, max_trie_entry_size - 1, -1):
		t += i

	assert len(t._partitions) == 1
	assert big_rep in t._subtrees
	assert len(t._subtrees[big_rep]) == 2 * max_trie_entry_size

	t += max_trie_entry_size - 1

	assert len(t._partitions) == 2
	assert big_rep in t._subtrees
	assert small_rep in t._subtrees
	assert len(t._subtrees[big_rep]) == max_trie_entry_size
	assert len(t._subtrees[small_rep]) == max_trie_entry_size + 1


@given(lists(valid_int_entry, min_size=0, max_size=max_trie_value, unique=True))
def test_iter(entries):
	t = YFastTrie(max_trie_entry_size)

	for entry in entries:
		t += entry

	entries = sorted(entries)

	for entry in t:
		assert entry == entries.pop(0)

	assert len(entries) == 0


class YFastStateMachine(RuleBasedStateMachine):
	def __init__(self):
		super(YFastStateMachine, self).__init__()
		self.t = YFastTrie(max_trie_entry_size)

	def teardown(self):
		values = list(chain.from_iterable(self.t._subtrees.values()))

		for val in values:
			self.t -= val

	@invariant()
	def valid_count(self):
		tree_sizes = map(lambda t: len(t), self.t._subtrees.values())
		assert len(self.t) == sum(tree_sizes)

	@invariant()
	def valid_min(self):
		if len(self.t) == 0:
			assert self.t.min is None
		elif len(self.t) == 1:
			assert self.t.min == self.t.max
		else:
			for value in chain.from_iterable(self.t._subtrees.values()):
				assert value is self.t.min or self.t.min < value

	@invariant()
	def valid_max(self):
		if len(self.t) == 0:
			assert self.t.max is None
		elif len(self.t) == 1:
			assert self.t.max == self.t.min
		else:
			for value in chain.from_iterable(self.t._subtrees.values()):
				assert value is self.t.min or self.t.min < value

	@invariant()
	def valid_representatives(self):
		rep = self.t._partitions.min_node
		while rep is not None:
			pred = rep.pred
			succ = rep.succ

			max(self.t._subtrees[rep.value]) <= rep.value

			if pred is not None:
				max(self.t._subtrees[pred.value]) < min(self.t._subtrees[rep.value])

			if succ is not None:
				min(self.t._subtrees[succ.value]) > max(self.t._subtrees[rep.value])

			rep = rep.succ

	@invariant()
	def valid_subtree_values(self):
		for rep in self.t._subtrees.keys():
			assert rep in self.t._partitions
			assert YFastTrie._calculate_representative(rep, max_trie_entry_size) == rep

	@rule(val=valid_trie_entry)
	def insert_value(self, val):
		self.t += val

	@rule(val=valid_trie_entry)
	def remove_value(self, val):
		if val not in self.t:
			try:
				self.t -= val
			except ValueError:
				pass
		else:
			self.t -= val

YFastStateMachine.TestCase.settings = settings(max_examples=50, deadline=None)
test_y_fast_trie = YFastStateMachine.TestCase



# encoding: utf-8

################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                            (C) 2020, Jeremy Brown                            #
#       Released under version 3.0 of the Non-Profit Open Source License       #
################################################################################

from io import open
from os.path import abspath, dirname, join

from py_fast_trie.x_fast import XFastTrie
from py_fast_trie.y_fast import YFastTrie

module_root = dirname(abspath(__file__))

with open(join(module_root, "VERSION"), encoding="utf-8") as version_file:
	__version__ = version_file.read().strip()
